#
# Provides a simple shell for simulating pi-calculus terms based on BPMN diagrams
# Copyright (C) 2007 Frank Puhlmann.
#

require 'piexecutor.rb'
require 'pianalyzer.rb'
require 'bpmn2pi.rb'
require 'bpmnview.rb'
require 'orjoin.rb'

class PiShell

	def initialize(agent)
		
		open(agent)
		
	end

	# Collects a list of agents contained in a given agent
	def collectAgentIdentifiers(beh,list)
		
		#puts "C " + beh.to_s 
		
		# Inspect cases
		case beh
	
			when Restriction, Tau, Input, Output:
				# Move on to successor
					collectAgentIdentifiers(beh.successor,list)
							
			when Summation, Composition:
				# Collect from every part
				beh.parts.each do |p|
					collectAgentIdentifiers(p,list)
				end
				
			when AgentRef:
				# Inspect only if not in list
				if !(list.include?(beh.agent))
					list << beh.agent
					collectAgentIdentifiers(beh.agent.behavior,list)
				end				
				
			when Agent:
				# Inspect only if not in list
				if !(list.include?(beh))
					list << beh
					collectAgentIdentifiers(beh.behavior,list)
				end
		end
	end
	
	def highlightNodes(active_nodes,blocked_nodes)
	
		if @bpmnView	
			@bpmnView.clearHighlights() 
			active_nodes.size.times do |n|									
				ident = 0						
				# 1. Collect id's
				case active_nodes[n]
					when IntraactionNodes:
						if active_nodes[n].input.value.function
							# Get node from function
							ident = active_nodes[n].input.value.function.to_s.sub(/N[0-9]*_/,'').to_i
							@bpmnView.highlight(ident, "orange")				
							ident = 0
							else
							ident = active_nodes[n].input.value.object.to_s.sub(/e/,'').to_i
							end
					when Node:
						ident = active_nodes[n].value.function.to_s.sub(/N[0-9]*_/,'').to_i
				end
				@bpmnView.highlight(ident, "green")
				end
			
			blocked_nodes.size.times do |n|
				ident = 0
				case blocked_nodes[n]
					when IntraactionNodes:
						ident = blocked_nodes[n].input.value.function.to_s.sub(/N[0-9]*_/,'').to_i
					end				
				@bpmnView.highlight(ident, "red")					
			end
				
			end
	end

	# Opens the shell
	def open(agent)
	
		# Define defaults for global options
		@options = Hash.new
		@options['forceCycles'] = false
	
	
		puts "Interactive pi-calculus shell (November, 22 2007)"
		puts "(c) 2007 Frank Puhlmann, puhlmann@hpi.uni-potsdam.de"
		
		parser = nil
		processes = Array.new
		sel = 0
		bpmnViewThread = nil
		@bpmnView = nil
		
		active_nodes = []
		blocked = []
		
		if agent
			exec = AsyncPiExecutor.new(agent) 
	        puts exec.currentAgent.to_definition if agent			
			active_nodes = exec.collectActivePrefixes if agent
			processes << agent
		end
		
		while true
		
			if active_nodes.size > 0
		
				#
				# Filter blocked nodes according to or-join semantics
				#
				active_nodes, blocked = quick_or_join(active_nodes, parser, sel)				
				
				if blocked.size > 0		
					puts "Blocked transitions:"
					blocked.each do |n|
						puts "-: " + n.to_s
						end	
					end				
					
				puts "Select transition:"
				active_nodes.size.times do |n|			
					puts n.to_s + ": " + active_nodes[n].to_s				
				end
			else					
				@bpmnView.clearHighlights() if @bpmnView
				puts "(No transitions available)"
			end
			
			cnt = false
			selection = -1
					
			while  (!cnt) do 
			
				# Highlight active nodes in BPMN view
				highlightNodes(active_nodes,blocked)
			
				print ">"
				$stdout.flush
				input = gets
				
				case input
				
					when "a\n", "agents\n":
						puts "AGENTS"
						
						processes.each do |p|
							idents = Array.new
							collectAgentIdentifiers(p,idents)
							idents.uniq!
							
							idents.each do |i|
								puts i.class.to_s + " " + i.to_definition
							end
						
						end
						
					when /^e |^export / 
					
    					file = input.sub(/^e |^export /, "").tr("\n","")
    					
    					puts "EXPORT TO " + file
    					
    					f = File.new(file,'w+')
    					
						processes.each do |p|
							idents = Array.new
							collectAgentIdentifiers(p,idents)
							idents.uniq!
							
							idents.each do |i|
								f << "agent " + i.to_definition + "\n"
							end   
						end 	

						f.close				
    					
						
					when "g\n", "graphics\n":
															
						print "GRAPHICS "
					
						if !parser or !parser.xml
							puts "NO FILE LOADED!"
							next
							end

						if bpmnViewThread
							# Kill view window
							@bpmnView.kill if @bpmnView
							@bpmnView = nil
							bpmnViewThread = nil
							puts "OFF"
							else
							# Create new view window
							bpmnViewThread = Thread.new do 
								@bpmnView = BPMNView.new(parser.xml,sel)	
								@bpmnView.run
								end
							puts "ON" 
							# Wait for Tk to initialize
							sleep 2
							end
					
					when "h\n", "help\n":
					
						puts "HELP"
						puts "a, agents: shows all agents"
						puts "[c, current <agent>: sets the current agent]"
						puts "[d, define <agent>: defines an agent]"
						puts "e, export <file>: exports current agents to file"
						puts "g, graphics: toggles the BPMN visualization of the current process on or off"
						puts "h, help: this help"
						puts "i, inspect <agent>: inspects the given agent"
						puts "l, load <file>: load file"
						puts "o, option <value key>: shows and toggles options"
						puts "q, quit: quit pi-shell"
						puts "r, reset: reset current agent"
						puts "s, states: calculates the remaining (different) states of the current agent"
						puts "t, tree: prints the current execution tree"
						puts "v, view: shows the given variable (type 'v' for help)"
						
					when /^i |^inspect /
					
						id = input.sub(/^i |^inspect /, "").tr("\n","")
						
						found = false
						
						processes.each do |p|
							idents = Array.new
							collectAgentIdentifiers(p,idents)
							idents.uniq!
												
							idents.each do |i| 
								if i.ident.to_s == id.to_s
									puts "INSPECT " + id.to_s + " " + i.description.to_s 
									puts i.class.to_s + " " + i.to_definition
									found = true
								end
							end
						end
						
						puts "INSPECT - " + id.to_s + " - NOT FOUND" if !found
						
					when /^l |^load /  
					
    					file = input.sub(/^l |^load /, "").tr("\n","")
    					
    					file = "omni-export.xml" if file == "" 			
						puts "LOAD " + file 	
						begin					
						parser = BPMN2PiParser.new(file,@options)
						processes = parser.processes
						if processes.size > 1
							puts "Select agent id for simulation (multiple processes found):"
							processes.size.times do |i| 
								puts "" + (i+1).to_s + ": " + processes[i].to_s + " " + 
											processes[i].agent.description.to_s
							end
							sel = -1
							while (sel<1 || sel>processes.size)
								print "?"
								sel = gets.to_i
							end
							agent = processes[sel-1]
						else
							sel = 1
							agent = processes[0]
						end
						
						exec = AsyncPiExecutor.new(agent) if agent					
						cnt = true
						
						# Update BPMN view if existing
						if @bpmnView
							@bpmnView.kill 
							@bpmnView = nil
							bpmnViewThread = nil
							end
						
					#	rescue
					#		puts "I/O ERROR"
							end	
							
					when /^o|^option /
					
						puts "OPTIONS"
					
						set = input.sub(/^o|^option/, "").tr("\n","").strip
						
						if set == "" 
							# Print current options
							@options.each_pair {|key, value| puts "#{key} = #{value}" }
															
							else
							# Set new option
							
							op, val = set.split('=')
							op.strip!; val.strip!
							
							puts "Setting #{op} to #{val}."
							
							@options[op]=val
							
							end
						

					when /^s|^states/ 
					
						puts "STATES"
					
					    max = input.sub(/^s |^states /, "").tr("\n","").to_i

						begin
						if agent
							analyzer = AsyncPiAnalyzer.new(exec.currentAgent,parser,sel)
							analyzer.maxStates = max if max > 0
							c = analyzer.remainingStates().to_i
							if c < analyzer.maxStates.to_i
								puts c.to_s
								else
								puts "More than " + analyzer.maxStates.to_s
								end
						else
							puts "0"
						end
						
						rescue SystemStackError
							puts "SYSTEM STACK ERROR"
						end
						
					when "t\n", "tree\n":
						
						puts "TREE"
						exec.currentAgent.to_tree if exec
						
										
					when "q\n", "quit\n":
					
						puts "QUIT"
					
						exit
						
					when "r\n", "reset\n"
					
						puts "RESET"
					
						exec = AsyncPiExecutor.new(agent) if agent
						cnt = true
						
					when /^v|^view/ 
					
					    var = input.sub(/^v |^view /, "").tr("\n","").to_s		
					    
					    case var
					    	when "sp":
						    	puts "VIEW SHORTEST PATHS"
					    		if parser
						    		parser.analyzers[sel-1].printMatrix(parser.analyzers[sel-1].tcMatrix)
						    		else
						    		puts "NO DIAGRAMS LOADED"
						    		end
						    when "xml":
						    	puts "VIEW XML"
						    	if parser
						    		puts parser.xml.to_s
						    		else
						    		puts "NO DIAGRAMS LOADED"
						    		end
						    when "nc":
						    	puts "VIEW NC"
						    	if parser
						    		puts parser.analyzers[sel-1].tcMatrix.size
						    		else
						    		puts "NO DIAGRAMS LOADED"
						    		end
						    else 
						    	puts "VIEW sp: shortest paths in the current BPMN diagram"
						    	puts "     xml: the currently loaded XML file"
						    	puts "     nc: the node count of the current process"
					    end
					    				
						
					when "\n"
					
						selection = 0
						cnt = true
						
					when /[0-9]+/
						
						selection = input.to_i if input.size>1
						if (selection>=0 and selection <active_nodes.size)
							cnt = true
						end		
						
					else 
					
						puts "Unkown/incomplete command (type 'h' for help)"
						
				end
			end
			
			# Execute transition here
			if selection >= 0 && agent
				puts "-->(" + selection.to_s + ") "
			
				exec.exec(active_nodes[selection])
			end
			
			# Get the next active nodes
			if agent
				puts exec.currentAgent.to_definition
				active_nodes = exec.collectActivePrefixes
				blocked = []
			end
			
		end		
		
	end
	
end

PiShell.new(nil)


